@@ -2,19 +2,6 @@ from functools import wraps
 from itertools import starmap, permutations
 import sys
 
-# Todo:
-#   replace config
-#   * don't overwrite existing namespace (e.g., function name assumed to be value)
-#   X fixed? * can't have multiple unspecified_default
-#   * dependency language (e.g. a -> b or c)
-#   * show relationships in -h
-#   * make app (vs. existing) args more obvious
-#   * editdist spell suggestions (e.g., ``did you mean X?")
-#   * required mismatches with unspecified_default if arg is not made explicit
-#   * error is --one-flag not specified even if multiple are allowed
-#   * deal with difference between localized/non
-#   * enumeration type
-
 
 def Config(filename, dictionary=None, overwrite=True):
     if dictionary is None:
@@ -29,14 +16,14 @@ def Config(filename, dictionary=None, overwrite=True):
             if dictionary.get(toks[0], None) is not None and not overwrite:
                 continue
             dictionary.__setitem__(*toks)
-    
+
     return dictionary
 
 
 # ---------- decorators ---------- #
 
-def names_to_options(f):
-    ''' Convert any option names to :class:`Option` objects. '''
+def options_to_names(f):
+    ''' Convert any :class:`Option`s to names. '''
 
     @wraps(f)
     def inner(*args, **kwargs):
@@ -60,7 +47,7 @@ def verify_args_exist(f):
 
         self = args[0]
         for arg in args[1:]:
-            if isinstance(arg, basestring) and arg not in self.options:
+            if isinstance(arg, basestring) and arg not in self._readers:
                 raise_error(arg)
 
         return f(*args, **kwargs)
@@ -99,19 +86,6 @@ def localize(f):
 # ---------- end decorators ---------- #
 
 
-class Multidict(dict):
-    class _MyList(list):
-        pass
-
-    def __setitem__(self, key, value):
-        v = self.get(key, None)
-        if v is None:
-            v = Multidict._MyList()
-            super(Multidict, self).__setitem__(key, v)
-
-        v.append(value)
-
-
 # ---------- exceptions ---------- #
 
 
@@ -135,7 +109,8 @@ class ManyAllowedNoneSpecifiedArgumentError(ArgumentError):
     e.g., one of -a and -b is required and neither is specified. '''
 
     def __init__(self, allowed):
-        super(ManyAllowedNoneSpecifiedArgumentError, self).__init__('[%s] not specified' % ', '.join(allowed))
+        super(ManyAllowedNoneSpecifiedArgumentError, self).__init__(('[%s] not'
+                + ' specified') % ', '.join(allowed))
 
 
 class UnspecifiedArgumentError(ArgumentError):
@@ -150,12 +125,21 @@ class MultipleSpecifiedArgumentError(ArgumentError):
 
 
 class DependencyError(ArgumentError):
-    ''' User specified an argument that requires another, unspecified argument. '''
+    ''' User specified an argument that requires another, unspecified argument.
+    '''
+
     pass
 
 
 class ConflictError(ArgumentError):
-    ''' User specified an argument that conflicts another specified argument. '''
+    ''' User specified an argument that conflicts another specified argument.
+    '''
+
+    pass
+
+
+class FailedConditionError(ArgumentError):
+    ''' Condition failed. '''
     pass
 
 # ---------- end exceptions ---------- #
@@ -174,18 +158,24 @@ class Option(object):
         ''' Specifiy other options which this argument requires.
 
         :param others: required options
-        :type others: sequence of either :class:`Option` or basestring of option names
+        :type others: sequence of either :class:`Option` or basestring of \
+                    option names
         '''
-        [self.parser.set_requires(self.argname, x) for x in others]
+        [self.parser._set_requires(self.argname, x) for x in others]
+        return self
+
+    def condition(self, func):
+        self.parser._set_condition(self.argname, func)
         return self
 
     def conflicts(self, *others):
         ''' Specifiy other options which this argument conflicts with.
 
         :param others: conflicting options
-        :type others: sequence of either :class:`Option` or basestring of option names
+        :type others: sequence of either :class:`Option` or basestring of \
+            option names
         '''
-        [self.parser.set_conflicts(self.argname, x) for x in others]
+        [self.parser._set_conflicts(self.argname, x) for x in others]
         return self
 
     def shorthand(self, alias):
@@ -194,7 +184,8 @@ class Option(object):
 
         >>> parser.str('option').shorthand('o')
 
-        would cause '--option' and '-o' to be alias argument labels when invoked on the command line.
+        would cause '--option' and '-o' to be alias argument labels when
+        invoked on the command line.
 
         :param alias: alias of argument
         '''
@@ -204,7 +195,7 @@ class Option(object):
 
     def default(self, value):
         ''' Provide a default value for this argument. '''
-        self.parser.set_default(self.argname, value)
+        self.parser._set_default(self.argname, value)
         return self
 
     def cast(self, cast):
@@ -218,6 +209,12 @@ class Option(object):
         self.parser._set_required(self.argname)
         return self
 
+    def unless(self, *replacements):
+        ''' Argument is required unless replacements specified. '''
+
+        self.parser._set_required(self.argname, replacements)
+        return self
+
     def unspecified_default(self):
         ''' Indicate that values passed without argument labels will be
         attributed to this argument. '''
@@ -227,14 +224,23 @@ class Option(object):
 
     def multiple(self):
         ''' Indicate that the argument can be specified multiple times. '''
-        self.parser.set_multiple(argname)
+        self.parser._set_multiple(self.argname)
+        return self
 
 # ---------- Argument readers ---------- #
 
+# Argument readers help parse the command line. The flow is as follows:
+#
+# 1) We create a reader for each argument based on the type specified by the
+#    programmer (e.g., a bool will get a _FlagArgumentReader
+#
+# 2) The reader is used when we hit the argument label
+
 
 class _ArgumentReader(object):
-    def __init__(self):
+    def __init__(self, parent):
         self.value = None
+        self.parent = parent
         self._init()
 
     def consume_or_skip(self, arg):
@@ -253,8 +259,9 @@ class _MultiWordArgumentReader(_ArgumentReader):
         self.value = []
 
     def consume_or_skip(self, arg):
-        if arg.startswith('-'):
+        if self.parent._is_argument_label(arg):
             return False
+
         self.value.append(arg)
         return True
 
@@ -300,20 +307,41 @@ class _SingleWordReader(_ArgumentReader):
 class Parser(object):
     ''' Command line parser. '''
 
-    def __init__(self, store=None, to_underscore=False, default_help=True):
-        self.options = {}
-        self.option_labels = {}
-        self._required = {}
-        self.multiple = set()
+    def __init__(self, store=None, default_help=True):
+        self._readers = {}
+        self._option_labels = {}
+        self._multiple = set()
         self._casts = {}
-        self.defaults = {}
-        self.extras = []
+        self._defaults = {}
+        self._extras = []
         self._unspecified_default = None
-        self.requires = Multidict()
         self.require_n = {}
-        self.conflicts = Multidict()
-        self.alias = {}
-        self._to_underscore = to_underscore
+
+        # dict of A (required) -> args that could replace A if A is not
+        # specified
+        self._required = {}
+
+        # dict of A -> args that A depends on
+        self._requires = {}
+
+        # dict of A -> args that A conflicts with
+        self._conflicts = {}
+        self._alias = {}
+
+        # convert arg labels to understore in value dict
+        self._to_underscore = False
+
+        # prefix to shorthand args
+        self._single_flag = '-'
+
+        # prefix to fullname args
+        self._double_flag = '--'
+
+        # function conditions
+        self._conditions = {}
+
+        # help message
+        self._help_prefix = None
 
         # set by user
         self._init_user_set(store)
@@ -321,41 +349,60 @@ class Parser(object):
         if default_help:
             self.flag('help').shorthand('h')
 
+    def help_prefix(self, message):
+        self._help_prefix = message
+        return self
+
+    def underscore(self):
+        self._to_underscore = True
+        return self
+
+    def single_flag(self, flag):
+        ''' Set the single flag prefix. This appears before short arguments
+        (e.g., -a). '''
+
+        if self._double_flag in flag:
+            raise ValueError('single_flag cannot be superset of double_flag')
+        self._single_flag = flag
+        return self
+
+    def double_flag(self, flag):
+        ''' Set the double flag prefix. This appears before long arguments
+        (e.g., --arg). '''
+
+        if flag in self._single_flag:
+            raise ValueError('single_flag cannot be superset of double_flag')
+
+        self._double_flag = flag
+        return self
+
     def _init_user_set(self, store=None):
-        self.user_values = Multidict()
-        self.specified = {}
+        self.user_values = {}
+        self._specified = {}
         if store is None:
             store = {}
-        self.store = store
+        self._store = store
         self._namemaps = {}
         self._rnamemaps = {}
 
-    @classmethod
-    def _get_locals_dict(cls):
-        # XXX need to get start frame, not just previous
-        pass
-
     @classmethod
     def with_locals(cls):
         ''' Create :class:`Parser` using locals() dict. '''
+
         import inspect
         vals = inspect.currentframe().f_back.f_locals
-#        p = Parser(cls._get_locals_dict())
-        p = Parser(vals)
-        p._to_underscore = True
-        return p
-
+        return Parser(vals).underscore()
 
     def _to_flag(self, name):
         name = self._unlocalize(name)
 
         if len(name) == 1:
-            return '-' + name
+            return self._single_flag + name
         else:
-            return '--' + name
+            return self._double_flag + name
 
-    @names_to_options
-    def set_mutually_required(self, *names):
+    @options_to_names
+    def require_all_if_any(self, *names):
         ''' All arguments require each other; i.e., if any is specified, then
         all must be specified. '''
 
@@ -365,10 +412,10 @@ class Parser(object):
                 if v == vi:
                     continue
 
-                self.set_requires(v, vi)
+                self._set_requires(v, vi)
 
-    @names_to_options
-    def set_at_least_one_required(self, *names):
+    @options_to_names
+    def require_at_least_one(self, *names):
         ''' At least one of the arguments is required. '''
 
         s = set(names)
@@ -383,14 +430,16 @@ class Parser(object):
         self._unspecified_default = name
 
     @localize
-    @names_to_options
+    @options_to_names
     def _set_required(self, name, replacements=None):
-        if replacements is None:
-            replacements = []
-        self._required[name] = replacements
+        self._required.setdefault(name, []).extend(replacements or [])
 
-    def add_config(self, name, **kwargs):
-        return self._add_option(name.cast(Config), **kwargs)
+    @localize
+    def _set_condition(self, name, condition):
+        self._conditions.setdefault(name, []).append(condition)
+
+    def config(self, name):
+        return self._add_option(name).cast(Config)
 
     def int(self, name):
         ''' Add integer argument. '''
@@ -398,33 +447,43 @@ class Parser(object):
 
     def float(self, name):
         ''' Add float argument. '''
-        return self._add_option(name.cast(float), **kwargs)
+        return self._add_option(name).cast(float)
 
     def str(self, name):
         ''' Add :py:class:`str` argument. '''
         return self._add_option(name)
 
-    def add_range(self, name):
+    def range(self, name):
+        ''' Range type, specified as python range. XXX '''
+
         def caster(x):
             def raise_error():
-                raise FormatError('%s is not range format: N:N+i ' % x)
+                raise FormatError(('%s is not range format: N:N+i, N-N+i, or N'
+                        + ' N+i') % x)
 
-            toks = x.split(':')
-            if len(toks) != 2:
+            for char in (' :-'):
+                if char in x:
+                    splitter = char
+                    break
+
+            toks = x.split(splitter)
+            if not (1 <= len(toks) <= 3):
                 raise_error()
             try:
                 return xrange(*(int(y) for y in toks))
             except ValueError:
                 raise_error()
 
-        return self._add_option(name).cast(caster)
+        result = self._add_option(name).cast(caster)
+        self._set_reader(name, _MultiWordArgumentReader)
+        return result
 
     def __getitem__(self, name):
         return Option(name, self)
 
     @localize
-    def _set_option(self, name, option):
-        self.options[name] = option
+    def _set_reader(self, name, option):
+        self._readers[name] = option
 
     def multiword(self, name):
         ''' Accepts multiple terms as an argument. For example:
@@ -433,7 +492,7 @@ class Parser(object):
         '''
 
         result = self._add_option(name)
-        self._set_option(name, _MultiWordArgumentReader)
+        self._set_reader(name, _MultiWordArgumentReader)
         return result
 
     def bool(self, name):
@@ -446,43 +505,42 @@ class Parser(object):
         while an absence indicates false. No arguments. '''
 
         result = self._add_option(name)
-        self._set_option(name, _FlagArgumentReader)
+        self._set_reader(name, _FlagArgumentReader)
         return result
 
     def _add_shorthand(self, source, alias):
-        if source not in self.options:
+        if source not in self._readers:
             raise ValueError('%s not an option' % source)
-        self.alias[alias] = source
+        if alias in self._alias:
+            raise ValueError('{0} already shorthand for {1}'.format(alias,
+                self._alias[alias]))
 
-    def _add_option(self, name,
-            argument_label=None, required=False, multiple=False,
-            unspecified_default=False):
+        self._alias[alias] = source
+
+    def _add_option(self, name, argument_label=None):
 
         name = self._localize(name)
 
-        if name in self.options:
+        if name in self._readers:
             raise ValueError('multiple types specified for %s' % name)
 
-        self.options[name] = _SingleWordReader
+        self._readers[name] = _SingleWordReader
 
         if argument_label is not None:
-            self.option_labels[name] = argument_label
-
-        if required:
-            self._set_required(name)
+            self._option_labels[name] = argument_label
 
         return Option(name, self)
 
     def _getoption(self, option):
-        o = self.options.get(option, None)
+        o = self._readers.get(option, None)
         if o is not None:
             return option, o
 
-        source = self.alias.get(option, None)
+        source = self._alias.get(option, None)
         if source is None:
             return option, None
 
-        return source, self.options.get(source, None)
+        return source, self._readers.get(source, None)
 
     def _localize(self, key):
         if self._to_underscore:
@@ -508,6 +566,10 @@ class Parser(object):
 
         return new_args
 
+    def _is_argument_label(self, arg):
+        return (arg.startswith(self._single_flag) or
+                arg.startswith(self._double_flag))
+
     def _read(self, args):
         argument_value = None
 
@@ -518,35 +580,33 @@ class Parser(object):
                 argument_value = None
 
             argument_name = None
-            has_label = arg.startswith('-')
 
-            if has_label:
-                short_cut = len(arg) == 2
-
-                if short_cut:
-                    arg = arg[1:]
+            if self._is_argument_label(arg):
+                if arg.startswith(self._double_flag):
+                    prefix = self._double_flag
                 else:
-                    if not arg.startswith('--'):
-                        raise UnspecifiedArgumentError(arg)
-                    arg = arg[2:]
+                    prefix = self._single_flag
+
+                arg = arg[len(prefix):]
 
                 argument_name = self._localize(arg)
 
                 argument_name, argument_value = self._getoption(argument_name)
                 if argument_value is None:
                     raise UnspecifiedArgumentError(argument_name)
-                argument_value = argument_value()
+                argument_value = argument_value(self)
             elif self._unspecified_default is not None:
                 argument_name = self._unspecified_default
 
                 # push value onto _SingleWordReader
-                argument_value = _SingleWordReader()
+                argument_value = _SingleWordReader(self)
                 argument_value.consume_or_skip(arg)
 
             if argument_name:
-                self.user_values[argument_name] = argument_value
+                self.user_values.setdefault(argument_name,
+                        []).append(argument_value)
             else:
-                self.extras.append(arg)
+                self._extras.append(arg)
 
     def _help_if_necessary(self):
         if 'help' in self.user_values:
@@ -555,13 +615,13 @@ class Parser(object):
 
     def _find_duplicates(self):
         for key, values in self.user_values.iteritems():
-            if len(values) > 1 and key not in self.multiple:
-                raise MultipleSpecifiedArgumentError('%s specified multiple times' %
-                        self._to_flag(key))
+            if len(values) > 1 and key not in self._multiple:
+                raise MultipleSpecifiedArgumentError(('%s specified multiple'
+                        + ' times') % self._to_flag(key))
 
     def _find_requirements(self):
         for key, values in self.user_values.iteritems():
-            for r in self.requires.get(key, []):
+            for r in self._requires.get(key, []):
                 if self._localize(r) not in self.user_values:
                     raise DependencyError('%s requires %s' %
                             (self._to_flag(key), self._to_flag(r)))
@@ -583,13 +643,12 @@ class Parser(object):
 
     def _find_conflicts(self):
         for key, values in self.user_values.iteritems():
-            for r in self.conflicts.get(key, []):
+            for r in self._conflicts.get(key, []):
                 if r in self.user_values:
                     raise ConflictError('%s conflicts with %s' %
                             (self._to_flag(key),
                              self._to_flag(r)))
 
-
     def _validate_entries(self):
         self._find_duplicates()
         self._find_requirements()
@@ -597,25 +656,31 @@ class Parser(object):
 
     def _assign(self):
         for key, values in self.user_values.iteritems():
-            try:
-                if key not in self.multiple:
-                    self._set_specified(key, values[0].get())
-                else:
-                    self._set_specified(key, [v.get() for v in values])
-            except ValueError:
-                raise MissingRequiredArgumentError('No value passed for %s' %
-                        self._unlocalize(key))
+            if key not in self._multiple:
+                self._set_specified(key, values[0].get())
+            else:
+                self._set_specified(key, [v.get() for v in values])
 
-        for key, value in self.options.iteritems():
-            if key not in self.specified:
+        # check conditions
+        for key, value in self._specified.iteritems():
+            val = self._conditions.get(key)
+            if not val:
+                continue
+
+            for cond in val:
+                if not cond(self._specified):
+                    raise FailedConditionError()
+
+        for key, value in self._readers.iteritems():
+            if key not in self._specified:
                 res = self._required.get(key, None)
                 if res is not None:
                     if len(res) == 0:
-                        raise MissingRequiredArgumentError('No value passed for'
-                                + ' %s' % self._unlocalize(key))
+                        raise MissingRequiredArgumentError('No value passed'
+                                + ' for' + ' %s' % self._unlocalize(key))
 
                     for result in res:
-                        if result in self.specified:
+                        if result in self._specified:
                             break
                     else:
                         raise ManyAllowedNoneSpecifiedArgumentError(
@@ -624,27 +689,34 @@ class Parser(object):
 #                else: raise MissingRequiredArgumentError('No value passed for
 #                %s' % key)
 
-        for key, value in self.specified.iteritems():
+        for key, value in self._specified.iteritems():
             self._set_final(key, value)
 
-        for key, value in self.options.iteritems():
-            if key in self.defaults:
-                self._set_final(key, self.defaults[key])
+        for key, value in self._readers.iteritems():
+            if key in self._defaults:
+                self._set_final(key, self._defaults[key])
             else:
                 self._set_final(key, value.default())
 
-    @names_to_options
+    @options_to_names
     def set_one_required(self, *names):
-        self.set_mutually_exclusive(*names)
-        self.set_at_least_one_required(*names)
+        self.mutually_exclude(*names)
+        self.require_at_least_one(*names)
 
-    @names_to_options
-    def set_mutually_dependent(self, *names):
-        list(starmap(self.set_requires, permutations(names, 2)))
+    def require_one(self, *names):
+        self.set_one_required(*names)
 
-    @names_to_options
-    def set_mutually_exclusive(self, *names):
-        list(starmap(self.set_conflicts, permutations(names, 2)))
+    @options_to_names
+    def all_require(self, required, *names):
+        [self._set_requires(name, required) for name in names]
+
+    @options_to_names
+    def mutually_require(self, *names):
+        list(starmap(self._set_requires, permutations(names, 2)))
+
+    @options_to_names
+    def mutually_exclude(self, *names):
+        list(starmap(self._set_conflicts, permutations(names, 2)))
 
     def _get_args(self, args):
         if args is None:
@@ -663,7 +735,7 @@ class Parser(object):
             self._help_if_necessary()
             self._validate_entries()
             self._assign()
-            store = self.store
+            store = self._store
         except ArgumentError as e:
             raise e
         finally:
@@ -683,56 +755,58 @@ class Parser(object):
         else:
             print e
         print 'usage: %s' % sys.argv[0],
-        for key in self.options:
+        for key in self._readers:
             print '[%s]' % self._label(key),
 
         print
         sys.exit(1)
 
     def _set_specified(self, key, value):
-        self.specified[key] = value
+        self._specified[key] = value
 
     def _set_final(self, key, value):
-        if key not in self.store:
+        if key not in self._store:
             cast = self._casts.get(key, None)
             if cast and value is not None:
                 if cast == Config:
-                    Config(value, self.store, overwrite=False)
+                    Config(value, self._store, overwrite=False)
                 else:
                     try:
                         value = cast(value)
+                    except FormatError:
+                        raise
                     except ValueError:
                         raise FormatError('Cannot cast %s to %s', value, cast)
 
-            self.store[key] = value
+            self._store[key] = value
 
     @localize
-    @names_to_options
-    def set_multiple(self, name):
-        self.multiple.add(name)
+    @options_to_names
+    def _set_multiple(self, name):
+        self._multiple.add(name)
 
     @localize
-    @names_to_options
-    def set_default(self, name, value):
-        self.defaults[name] = value
+    @options_to_names
+    def _set_default(self, name, value):
+        self._defaults[name] = value
 
     @localize_all
     @verify_args_exist
-    @names_to_options
-    def set_requires(self, a, b):
-        self.requires[a] = b
+    @options_to_names
+    def _set_requires(self, a, b):
+        self._requires.setdefault(a, []).append(b)
 
     @localize_all
     @verify_args_exist
-    @names_to_options
+    @options_to_names
     def set_requires_n_of(self, a, n, *others):
         self.require_n[a] = (n, others)
 
     @localize_all
     @verify_args_exist
-    @names_to_options
-    def set_conflicts(self, a, b):
-        self.conflicts[a] = b
+    @options_to_names
+    def _set_conflicts(self, a, b):
+        self._conflicts.setdefault(a, []).append(b)
 
     def __enter__(self):
         return self
@@ -741,35 +815,38 @@ class Parser(object):
         self.process_command_line()
 
     def _option_label(self, name):
-        return self.option_labels.get(name, 'option')
+        return self._option_labels.get(name, 'option')
 
     def _label(self, name):
         pkey = self._to_flag(name)
-        for k, v in self.alias.iteritems():
+        for k, v in self._alias.iteritems():
             if name in v:
                 pkey += ',' + self._to_flag(k)
 
-        value = self.options[name]
+        value = self._readers[name]
         if value != _FlagArgumentReader:
             pkey = '%s <%s>' % (pkey, self._option_label(name))
 
         return pkey
 
     def print_help(self):
+        if self._help_prefix:
+            print self._help_prefix
+
         print 'Arguments:'
         column_width = -1
-        for key in self.options:
+        for key in self._readers:
             column_width = max(column_width, len(self._label(key)))
 
         fmt = '   %-' + str(column_width) + 's'
 
-        for key in self.options:
+        for key in self._readers:
             print fmt % self._label(key)
 
 
 class IOParser(Parser):
     def directory(self, name):
-        ''' Adds a directory option, as :class:`io.Directory` object.  Requires
+        ''' Adds a directory option, as :class:`io.Directory` object. Requires
         directory to exist or raises exception. '''
 
         from plyny.plio.files import Directory
@@ -781,7 +858,7 @@ class IOParser(Parser):
         return self._add_option(name).cast(open_path)
 
     def input_file(self, name):
-        ''' Adds an input file option, as :class:`io.File` object.  Requires
+        ''' Adds an input file option, as :class:`io.File` object. Requires
         file to exist or raises exception. '''
 
         from plyny.plio.files import open_path
@@ -792,7 +869,7 @@ class IOParser(Parser):
             return f
 
         result = self._add_option(name).cast(opener)
-        self._set_option(name, _MultiWordArgumentReader)
+        self._set_reader(name, _MultiWordArgumentReader)
         return result
 
     def output_file(self, name, disable_overwrite=True):
